==> python_string_fuzzy_clustering-0.0.1-py3-none-any.whl: sgfcmed/sgfcmed_parallel.py <==
"""
SGFCMedParallel: String Grammar Fuzzy C-Medians with multiprocessing
Author: Atcharin Klomsae
Date: 23/4/2020

Description:
This implementation performs fuzzy clustering on strings using Levenshtein distance
and modified median string updates. It uses multiprocessing to improve speed on large datasets.
"""

import random
import multiprocessing
from functools import lru_cache, partial

def _lev(s1, s2):
    # Initialize distance matrix
    len_s1, len_s2 = len(s1), len(s2)
    dp = [[0] * (len_s2 + 1) for _ in range(len_s1 + 1)]

    # Base cases: distance from empty string
    for i in range(len_s1 + 1): dp[i][0] = i
    for j in range(len_s2 + 1): dp[0][j] = j

    # Fill in the matrix
    for i in range(1, len_s1 + 1):
        for j in range(1, len_s2 + 1):
            cost = 0 if s1[i - 1] == s2[j - 1] else 1
            dp[i][j] = min(
                dp[i - 1][j] + 1,         # deletion
                dp[i][j - 1] + 1,         # insertion
                dp[i - 1][j - 1] + cost   # substitution
            )
           
    return dp[len_s1][len_s2]

def _parallel_total_distance(candidate, S):
        """
        Wrapper for multiprocessing: returns candidate and its total distance to S.
        """
        return candidate, sum(_lev(candidate, s) for s in S)

class SGFCMedParallel:
    def __init__(self, C=2, m=2.0, max_iter=100, tol=1e-4):
        """
        Initialize clustering model.

        Parameters:
            C (int): Number of clusters
            m (float): Fuzzifier (greater than 1)
            max_iter (int): Maximum number of iterations
            tol (float): Tolerance for convergence
        """
        self.C = C  # number of clusters
        self.m = m  # fuzzifier for controlling fuzziness
        self.max_iter = max_iter  # maximum number of training iterations
        self.tol = tol  # convergence threshold
        self.prototypes_ = []  # list of prototype strings (cluster centers)
        self.U_ = []  # fuzzy membership matrix
        self.__debug_candidates_per_cluster = {}

        random.seed(42)  # ensure reproducible random initialization

        # Prepare Levenshtein distance function with memoization
        self._levenshtein = self._memoized_levenshtein 


    @lru_cache(maxsize=None) # helps in reducing the execution time of the function by using memoization technique.
    def _memoized_levenshtein(self,s1,s2):
        """
        Create a memoized version of Levenshtein distance to avoid redundant computation.
        """
        return _lev(s1, s2)

    def _total_distance(self, candidate, S):
        """
        Compute the total distance from a candidate string to all strings in S.
        Used for finding the best prototype.
        """
        return sum(self._levenshtein(candidate, s) for s in S)

    def _update_prototype(self, s, S, alphabet, cluster_idx=None):
        all_candidates_per_round = []
        for i in range(len(s)):
            candidates = [s]

            for a in alphabet:
                if i < len(s) and a != s[i]:
                    candidates.append(s[:i] + a + s[i+1:]) 

            if len(s) > 1:
                candidates.append(s[:i] + s[i+1:]) 
                
            for a in alphabet:
                candidates.append(s[:i] + a + s[i:]) 

            all_candidates_per_round.append(candidates.copy())

        
            func = partial(_parallel_total_distance, S=S) # def func(candidate): return parallel_total_distance(candidate, S=S)
            with multiprocessing.Pool() as pool: 
                results = pool.map(func, candidates) 
                
            s = min(results, key=lambda x: x[1])[0] 
            if cluster_idx is not None:
                self.__debug_candidates_per_cluster[cluster_idx] = all_candidates_per_round
            

        return s

    def fit(self, S):
        """
        Train the model on list of strings S using fuzzy c-means clustering.
        """
        N = len(S)  # number of strings
        alphabet = set(''.join(S))  # all unique characters in the dataset

        # Step 1: Initialize fuzzy membership matrix randomly
        self.U_ = [[random.random() for _ in range(self.C)] for _ in range(N)] 
        for i in range(N):
            total = sum(self.U_[i]) 
            self.U_[i] = [u / total for u in self.U_[i]]  

        # Step 2: Initialize cluster prototypes randomly from input strings
        self.prototypes_ = random.sample(S, self.C)

        for iteration in range(self.max_iter):
            old_prototypes = self.prototypes_[:]

            # Step 3: Update cluster prototypes using modified median
            for i in range(self.C):
                self.prototypes_[i] = self._update_prototype(self.prototypes_[i], S, alphabet,cluster_idx=i)
                

            # Step 4: Update fuzzy membership values based on distance to prototypes
            for k in range(N):  # for each string
                for i in range(self.C):  # for each cluster
                    d_i = self._levenshtein(S[k], self.prototypes_[i]) + 1e-6  # avoid division by 0
                    denom = sum(
                        (d_i / (self._levenshtein(S[k], self.prototypes_[j]) + 1e-6)) ** (2 / (self.m - 1))
                        for j in range(self.C)
                    )
                    self.U_[k][i] = 1 / (denom + 1e-6)

            # Step 5: Check for convergence (change in prototypes)
            change = sum(self._levenshtein(p1, p2) for p1, p2 in zip(old_prototypes, self.prototypes_))
            # print("change ",change)
            if change < self.tol:
                break

    def membership(self):
        """
        Return the current fuzzy membership matrix.
        """
        return self.U_

    def prototypes(self):
        """
        Return the current list of cluster prototype strings.
        """
        return self.prototypes_

    def predict(self, S):
        """
        Assign each string in S to the nearest prototype (crisp clustering).
        """
        preds = []
        for s in S:
            distances = [self._levenshtein(s, proto) for proto in self.prototypes_]
            preds.append(distances.index(min(distances)))  # cluster with min distance
        return preds
    
    def get_debug_candidates(self, cluster_idx=None):
        """
        Return generated candidate prototypes for testing/debugging.
        If cluster_idx is given, return candidates for that cluster only.
        """
        if cluster_idx is not None:
            return self.__debug_candidates_per_cluster.get(cluster_idx, [])
        return self.__debug_candidates_per_cluster

==> python_string_fuzzy_clustering-0.0.3-py3-none-any.whl: sgfcmed/sgfcmed_parallel.py <==
"""
SGFCMedParallel: String Grammar Fuzzy C-Medians with multiprocessing
Author: Atcharin Klomsae
Date: 23/4/2020

Description:
This implementation performs fuzzy clustering on strings using Levenshtein distance
and modified median string updates. It uses multiprocessing to improve speed on large datasets.
"""

import random
import multiprocessing
from functools import lru_cache, partial

def _lev(s1, s2):
    # Initialize distance matrix
    len_s1, len_s2 = len(s1), len(s2)
    dp = [[0] * (len_s2 + 1) for _ in range(len_s1 + 1)]

    # Base cases: distance from empty string
    for i in range(len_s1 + 1): dp[i][0] = i
    for j in range(len_s2 + 1): dp[0][j] = j

    # Fill in the matrix
    for i in range(1, len_s1 + 1):
        for j in range(1, len_s2 + 1):
            cost = 0 if s1[i - 1] == s2[j - 1] else 1
            dp[i][j] = min(
                dp[i - 1][j] + 1,         # deletion
                dp[i][j - 1] + 1,         # insertion
                dp[i - 1][j - 1] + cost   # substitution
            )
           
    return dp[len_s1][len_s2]

def _parallel_total_distance(candidate, S):
        """
        Wrapper for multiprocessing: returns candidate and its total distance to S.
        """
        return candidate, sum(_lev(candidate, s) for s in S)

class SGFCMedParallel:
    def __init__(self, C=2, m=2.0, max_iter=100, tol=1e-4):
        """
        Initialize clustering model.

        Parameters:
            C (int): Number of clusters
            m (float): Fuzzifier (greater than 1)
            max_iter (int): Maximum number of iterations
            tol (float): Tolerance for convergence
        """
        self.C = C  # number of clusters
        self.m = m  # fuzzifier for controlling fuzziness
        self.max_iter = max_iter  # maximum number of training iterations
        self.tol = tol  # convergence threshold
        self.prototypes_ = []  # list of prototype strings (cluster centers)
        self.U_ = []  # fuzzy membership matrix
        self.__debug_candidates_per_cluster = {}

        random.seed(42)  # ensure reproducible random initialization

        # Prepare Levenshtein distance function with memoization
        self._levenshtein = self._memoized_levenshtein 


    @lru_cache(maxsize=None) # helps in reducing the execution time of the function by using memoization technique.
    def _memoized_levenshtein(self,s1,s2):
        """
        Create a memoized version of Levenshtein distance to avoid redundant computation.
        """
        return _lev(s1, s2)

    def _total_distance(self, candidate, S):
        """
        Compute the total distance from a candidate string to all strings in S.
        Used for finding the best prototype.
        """
        return sum(self._levenshtein(candidate, s) for s in S)

    def _update_prototype(self, s, S, alphabet, cluster_idx=None):
        all_candidates_per_round = []
        for i in range(len(s)):
            candidates = [s]

            for a in alphabet:
                if i < len(s) and a != s[i]:
                    candidates.append(s[:i] + a + s[i+1:]) 

            if len(s) > 1:
                candidates.append(s[:i] + s[i+1:]) 
                
            for a in alphabet:
                candidates.append(s[:i] + a + s[i:]) 

            all_candidates_per_round.append(candidates.copy())

        
            func = partial(_parallel_total_distance, S=S) # def func(candidate): return parallel_total_distance(candidate, S=S)
            with multiprocessing.Pool() as pool: 
                results = pool.map(func, candidates) 
                
            s = min(results, key=lambda x: x[1])[0] 
            if cluster_idx is not None:
                self.__debug_candidates_per_cluster[cluster_idx] = all_candidates_per_round
            

        return s

    def fit(self, S):
        """
        Train the model on list of strings S using fuzzy c-means clustering.
        """
        N = len(S)  # number of strings
        alphabet = set(''.join(S))  # all unique characters in the dataset

        # Step 1: Initialize fuzzy membership matrix randomly
        self.U_ = [[random.random() for _ in range(self.C)] for _ in range(N)] 
        for i in range(N):
            total = sum(self.U_[i]) 
            self.U_[i] = [u / total for u in self.U_[i]]  

        # Step 2: Initialize cluster prototypes randomly from input strings
        self.prototypes_ = random.sample(S, self.C)

        for iteration in range(self.max_iter):
            old_prototypes = self.prototypes_[:]

            # Step 3: Update cluster prototypes using modified median
            for i in range(self.C):
                self.prototypes_[i] = self._update_prototype(self.prototypes_[i], S, alphabet,cluster_idx=i)
                

            # Step 4: Update fuzzy membership values based on distance to prototypes
            for k in range(N):  # for each string
                for i in range(self.C):  # for each cluster
                    d_i = self._levenshtein(S[k], self.prototypes_[i]) + 1e-6  # avoid division by 0
                    denom = sum(
                        (d_i / (self._levenshtein(S[k], self.prototypes_[j]) + 1e-6)) ** (2 / (self.m - 1))
                        for j in range(self.C)
                    )
                    self.U_[k][i] = 1 / (denom + 1e-6)

            # Step 5: Check for convergence (change in prototypes)
            change = sum(self._levenshtein(p1, p2) for p1, p2 in zip(old_prototypes, self.prototypes_))
            # print("change ",change)
            if change < self.tol:
                break

    def membership(self):
        """
        Return the current fuzzy membership matrix.
        """
        return self.U_

    def prototypes(self):
        """
        Return the current list of cluster prototype strings.
        """
        return self.prototypes_

    def predict(self, S):
        """
        Assign each string in S to the nearest prototype (crisp clustering).
        """
        preds = []
        for s in S:
            distances = [self._levenshtein(s, proto) for proto in self.prototypes_]
            preds.append(distances.index(min(distances)))  # cluster with min distance
        return preds
    
    def _get_debug_candidates(self, cluster_idx=None):
        """
        Return generated candidate prototypes for testing/debugging.
        If cluster_idx is given, return candidates for that cluster only.
        """
        if cluster_idx is not None:
            return self.__debug_candidates_per_cluster.get(cluster_idx, [])
        return self.__debug_candidates_per_cluster
